@@ -1,4 +1,4 @@
-# Copyright (c) 2012 Erik Johansson <erik@ejohansson.se>
+# Copyright (c) 2012-2013 Erik Johansson <erik@ejohansson.se>
 # 
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU General Public License as
@@ -18,6 +18,7 @@
 from ctypes import c_bool, c_char_p, c_int, c_ubyte, c_ulong, c_void_p
 from ctypes import byref, cast, create_string_buffer, POINTER, sizeof
 import platform
+import threading
 
 
 if platform.system() == 'Windows':
@@ -40,29 +41,149 @@ CONTROLLER_EVENT_FUNC = FUNCTYPE(
 
 
 class TelldusError(Exception):
-    """Error returned from Telldus API.
+    """Error returned from Telldus Core API.
+
+    Automatically raised when a function in the C API returns an error.
+
+    Attributes:
+        error: The error code constant (one of TELLSTICK_ERROR_* from
+        :mod:`tellcore.constants`).
     """
-    def __init__(self, error):
+
+    def __init__(self, error, lib=None):
         super(TelldusError, self).__init__()
         self.error = error
+        self.lib = Library() if lib is None else lib
 
     def __str__(self):
-        msg = Library().tdGetErrorString(self.error)
+        """Return the human readable error string."""
+        msg = self.lib.tdGetErrorString(self.error)
         return "%s (%d)" % (msg, self.error)
 
 
+class BaseCallbackDispatcher(object):
+    """Base callback dispatcher class.
+
+    Inherit from this class and override the :func:`on_callback` method to
+    change how callbacks are dispatched.
+    """
+
+    def on_callback(self, callback, *args):
+        """Called from the callback thread when an event is received.
+
+        :param callable callback: The callback function to call.
+        :param args: The arguments to pass to the callback.
+        """
+        pass
+
+
+class DirectCallbackDispatcher(BaseCallbackDispatcher):
+    """Dispatches callbacks directly.
+
+    This is the default callback dispatcher when using the Library class
+    directly. Since the callback is dispatched directly, the callback is called
+    in the callback thread.
+
+    The recommended way is to use the :class:`tellcore.telldus.TelldusCore`
+    class instead, in which case the default dispatcher takes care of
+    dispatching the callback in the main thread.
+    """
+
+    def on_callback(self, callback, *args):
+        callback(*args)
+
+
 class Library(object):
+    """Wrapper around the Telldus Core C API.
+
+    With the exception of tdInit, tdClose and tdReleaseString, all functions in
+    the C API (see `Telldus Core documentation
+    <http://developer.telldus.com/doxygen/group__core.html>`_) can be
+    called. The parameters are the same as in the C API documentation. The
+    return value are mostly the same as for the C API, except for functions
+    with multiple out parameters.
+
+    The class takes care of:
+       * freeing memory for strings returned from the C API,
+       * converting an error return from a function into an exception
+         (TelldusError),
+       * converts to and from Python strings <=> C style strings.
+    """
+
     STRING_ENCODING = 'utf-8'
 
     class c_string_p(c_char_p):
-        def __init__(self, string):
-            if type(string) is str or type(string) is unicode:
-                string = string.encode(Library.STRING_ENCODING)
-            c_char_p.__init__(self, string)
+        def __init__(self, param):
+            c_char_p.__init__(self, param.encode(Library.STRING_ENCODING))
+
+        @classmethod
+        def from_param(cls, param):
+            if type(param) is str:
+                return cls(param)
+            try:
+                if type(param) is unicode:
+                    return cls(param)
+            except NameError:
+                pass # The unicode type does not exist in python 3
+            return c_char_p.from_param(param)
+
+    # Must be a separate class (i.e. not part of Library), to avoid circular
+    # references when saving the wrapper callback function in a class with a
+    # destructor, as the destructor is not called in that case.
+    class CallbackWrapper(object):
+        def __init__(self):
+            self._callbacks = {}
+            self._lock = threading.Lock()
+            self._dispatcher = DirectCallbackDispatcher()
+
+        def set_callback_dispatcher(self, dispatcher):
+            with self._lock:
+                self._dispatcher = dispatcher
+
+        def get_callback_ids(self):
+            with self._lock:
+                return list(self._callbacks.keys())
+
+        def register_callback(self, registrator, functype, callback):
+            wrapper = functype(self._callback)
+            with self._lock:
+                id = registrator(wrapper, None)
+                self._callbacks[id] = (wrapper, callback)
+                return id
+
+        def unregister_callback(self, id):
+            with self._lock:
+                del self._callbacks[id]
+
+        def _callback(self, *in_args):
+            args = []
+            # Convert all char* parameters (i.e. bytes) to proper python
+            # strings
+            for arg in in_args:
+                if type(arg) is bytes:
+                    args.append(arg.decode(Library.STRING_ENCODING))
+                else:
+                    args.append(arg)
+
+            # Get the real callback and the dispatcher
+            with self._lock:
+                try:
+                    # args[-2] is callback id
+                    (wrapper, callback) = self._callbacks[args[-2]]
+                except KeyError:
+                    return
+                dispatcher = self._dispatcher
+
+            # Dispatch the callback, dropping the last parameter which is the
+            # context and always None.
+            try:
+                dispatcher.on_callback(callback, *args[:-1])
+            except:
+                pass
 
     _lib = None
     _refcount = 0
-    _callbacks = {}
+    _callback_wrapper = CallbackWrapper()
 
     _functions = {
         'tdInit': [None, []],
@@ -145,7 +266,7 @@ class Library(object):
                 string = string.decode(Library.STRING_ENCODING)
             return string
 
-        for name, signature in self._functions.items():
+        for name, signature in Library._functions.items():
             try:
                 func = getattr(lib, name)
                 func.restype = signature[0]
@@ -165,6 +286,9 @@ class Library(object):
 
         The library is only initialized the first time this object is
         created. Subsequent instances uses the same library instance.
+
+        :param str name: Default value is the platform specific name of the
+            Telldus library, but it can be e.g. an absolute path.
         """
         super(Library, self).__init__()
 
@@ -192,18 +316,27 @@ class Library(object):
         assert Library._refcount >= 1
         Library._refcount -= 1
 
-        if Library._refcount == 0:
-            for callback in list(self._callbacks.keys()):
-                try:
-                    self.tdUnregisterCallback(callback)
-                except:
-                    pass
+        if Library._refcount != 0:
+            return
 
-            Library._lib.tdClose()
-            Library._lib = None
+        for id in Library._callback_wrapper.get_callback_ids():
+            try:
+                self.tdUnregisterCallback(id)
+            except:
+                pass
+
+        Library._lib.tdClose()
+        Library._lib = None
+
+    def set_callback_dispatcher(self, dispatcher):
+        """Change the callback dispatcher.
+
+        See documentation for :class:`BaseCallbackDispatcher`.
+        """
+        Library._callback_wrapper.set_callback_dispatcher(dispatcher)
 
     def __getattr__(self, name):
-        if name in self._functions:
+        if name in Library._functions:
             return getattr(self._lib, name)
         raise AttributeError(name)
 
@@ -217,40 +350,37 @@ class Library(object):
         raise NotImplementedError('should not be called explicitly')
 
     def tdRegisterDeviceEvent(self, callback):
-        func = DEVICE_EVENT_FUNC(callback)
-        id = self._lib.tdRegisterDeviceEvent(func, None)
-        self._callbacks[id] = func
-        return id
+        return Library._callback_wrapper.register_callback(
+            self._lib.tdRegisterDeviceEvent, DEVICE_EVENT_FUNC, callback)
 
     def tdRegisterDeviceChangeEvent(self, callback):
-        func = DEVICE_CHANGE_EVENT_FUNC(callback)
-        id = self._lib.tdRegisterDeviceChangeEvent(func, None)
-        self._callbacks[id] = func
-        return id
+        return Library._callback_wrapper.register_callback(
+            self._lib.tdRegisterDeviceChangeEvent, DEVICE_CHANGE_EVENT_FUNC,
+            callback)
 
     def tdRegisterRawDeviceEvent(self, callback):
-        func = RAW_DEVICE_EVENT_FUNC(callback)
-        id = self._lib.tdRegisterRawDeviceEvent(func, None)
-        self._callbacks[id] = func
-        return id
+        return Library._callback_wrapper.register_callback(
+            self._lib.tdRegisterRawDeviceEvent, RAW_DEVICE_EVENT_FUNC,
+            callback)
 
     def tdRegisterSensorEvent(self, callback):
-        func = SENSOR_EVENT_FUNC(callback)
-        id = self._lib.tdRegisterSensorEvent(func, None)
-        self._callbacks[id] = func
-        return id
+        return Library._callback_wrapper.register_callback(
+            self._lib.tdRegisterSensorEvent, SENSOR_EVENT_FUNC, callback)
 
     def tdRegisterControllerEvent(self, callback):
-        func = CONTROLLER_EVENT_FUNC(callback)
-        id = self._lib.tdRegisterControllerEvent(func, None)
-        self._callbacks[id] = func
-        return id
+        return Library._callback_wrapper.register_callback(
+            self._lib.tdRegisterControllerEvent, CONTROLLER_EVENT_FUNC,
+            callback)
 
     def tdUnregisterCallback(self, id):
-        del self._callbacks[id]
+        Library._callback_wrapper.unregister_callback(id)
         self._lib.tdUnregisterCallback(id)
 
     def tdSensor(self):
+        """Get the next sensor while iterating.
+
+        :return: a dict with the keys: protocol, model, id, datatypes.
+        """
         protocol = create_string_buffer(20)
         model = create_string_buffer(20)
         id = c_int()
@@ -263,6 +393,10 @@ class Library(object):
                 'id': id.value, 'datatypes': datatypes.value}
 
     def tdSensorValue(self, protocol, model, id, datatype):
+        """Get the sensor value for a given sensor.
+
+        :return: a dict with the keys: value, timestamp.
+        """
         value = create_string_buffer(20)
         timestamp = c_int()
 
@@ -271,6 +405,10 @@ class Library(object):
         return {'value': self._to_str(value), 'timestamp': timestamp.value}
 
     def tdController(self):
+        """Get the next controller while iterating.
+
+        :return: a dict with the keys: id, type, name, available.
+        """
         id = c_int()
         type = c_int()
         name = create_string_buffer(255)
